/*
 * ArrayMD class for multidimensional arrays
 *
 * Rahulkumar Gayatri
 * */
#ifndef _ARRAYMD_H
#define _ARRAYMD_H

#include <array>
#include <cassert>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

template<size_t dim, typename T>
inline void
compute_offsets(std::array<size_t, dim>& offsets, size_t& size, T arg)
{
  offsets[dim - 1] = arg;
  size *= arg;
}

template<size_t dim, typename T, typename... Idx>
inline void
compute_offsets(std::array<size_t, dim>& offsets,
                size_t& size,
                T arg,
                Idx... args)
{
  offsets[dim - (sizeof...(args)) - 1] = arg;
  size *= arg;
  compute_offsets(offsets, size, args...);
}

template<size_t dim, typename T>
inline void
compute_args(std::array<size_t, dim>& offsets, size_t& index, T arg)
{
  index += arg;
}

template<size_t dim, typename T, typename... Idx>
inline void
compute_args(std::array<size_t, dim>& offsets,
             size_t& index,
             T arg,
             Idx... args)
{

  auto index_eval = arg;
  for (int i = dim - sizeof...(args); i < dim; ++i)
    index_eval *= offsets[i];
  index += index_eval;

  compute_args(offsets, index, args...);

  return;
}

template<typename T, size_t dim>
struct ArrayMD
{
  std::array<size_t, dim> m_offsets;
  size_t size;
  T* dptr;

  ArrayMD() = default;

  template<typename... Idx>
  ArrayMD(Idx... args)
  {
    const auto N = sizeof...(args);

    // Check if the number of arguments passed matches the dimensionality passed in the template
    static_assert(N == dim,
                  "Dimensionality passed does not match the argument list");
    size = 1;
    m_offsets.fill(0);

    compute_offsets(m_offsets, size, args...);

    // Allocate memory for size number of T-elements
    dptr = new T[size];
  }

  template<typename... Idx>
  inline T& operator()(Idx... args)
  {
    const auto N = sizeof...(args);
    static_assert(N == dim, "parameters passed exceed the dimensionality");
    size_t index = 0;

    compute_args(m_offsets, index, args...);

    return dptr[index];
  }

  // Copy constructor sets size to 0 in order to avoid unintentional free-ing of the memory
  ArrayMD(const ArrayMD& p)
  {
    m_offsets = p.m_offsets;
    dptr = p.dptr;
    size = 0;
  }

  // Destructor
  ~ArrayMD()
  {
    if (size && dptr)
      delete[] dptr;
  }
};

template<typename T>
struct Array2D
{
  unsigned n1, n2, b1;
  unsigned size;
  T* dptr;

  // Calculate the index in the 2D array in row-major format.
  inline T& operator()(unsigned i1, unsigned i2)
  {
    return dptr[i2 + (n2 * i1)];
  }

  Array2D() = default;

  // Copy constructor sets size to 0 in order to avoid unintentional free-ing of the memory
  Array2D(const Array2D& p)
  {
    n1 = p.n1;
    n2 = p.n2;
    size = 0;
    dptr = p.dptr;
  }

  // Constructor allocates pointer to in1*in2 elements elements
  Array2D(int in1, int in2)
  {
    n1 = in1;
    n2 = in2;
    size = n1 * n2;
    dptr = new T[size];
  }

  // Destructor
  ~Array2D()
  {
    if (size && dptr)
      delete[] dptr;
  }
};
#endif
